unit XLS;

{
 Native Excel Writer (c)2003-2005, by Paul TOTH <tothpaul@free.fr>
 http://tothpaul.free.fr
}

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}


// based on
//
//  http://sc.openoffice.org/excelfileformat.pdf
//
// and
//
// ****************************************************************************
// * TmxNativeExcel v1.23
// * Native Excel Export component for Delphi.
// * Date last modified: 22.10.2002
// * Copyright 2001-2002, Bitvadsz Kft. All Rights Reserved.
// * Web page: www.maxcomponents.net
// ****************************************************************************

interface

uses
 Classes,Graphics,SysUtils;

const
 fmCurrency=7;

type
 TCellAlignment=(caGeneral,caLeft,caCenter,caRight,caFill);
 TCellBorder   =(cbLeft,cbRight,cbTop,cbBottom);
 TCellBorders  =set of TCellBorder;

 TExcelStream=class(TFileStream)
 private
  fFont:integer;
  fAlignment:TCellAlignment;
  fBorders  :TCellBorders;
  function GetFont:integer;
  procedure SetFont(Value:integer);
 public
  constructor Create(FileName:string);
  destructor Destroy; override;
  procedure AddFont(Font:TFont);
  procedure AddFormats;
  procedure SetColWidth(FirstCol, LastCol, Width:word);
  procedure AddInteger(x,y,Value:integer);
  procedure AddFloat(x,y:integer; Value:double);
  procedure AddCurrency(x,y:integer; Value:double);
  procedure AddString(x,y:integer; Value:string);
  procedure AddSum(x,y,c1,l1,c2,l2:integer; fmt:integer=0);
  procedure AddSumDiv2(x,y,c1,l1,c2,l2:integer; fmt:integer=0);
  procedure AddSumOfLines(x,y:integer; cells:array of integer; fmt:integer=0);
  procedure AddSubLines(x,y,y1,y2:integer; fmt:integer=0);
  procedure AddPercent(x,y,c1,l1,c2,l2:integer);
  procedure AddRef(x,y,c1,l1:integer; fmt:integer=0);
  property Font:integer read GetFont write SetFont;
  property Alignment:TCellAlignment read fAlignment write fAlignment;
  property Borders:TCellBorders read fBorders write fBorders;
 end;

implementation

// BIFF

type
 TBIFF_FONT=packed record
  Sign   : word;
  Len    : word;
  Height : word;
  Attr   : word;
  NameLen: byte;
 // FontName
 end;

 TBIFF_COLWIDTH=packed record
  Sign    : word;
  Len     : word;
  FirstCol: byte;
  LastCol : byte;
  Width   : word;
 end;

 TBIFF_CELL=packed record
  Sign  : word;
  Len   : word;
  Row   : word;
  Col   : word;
  Prot  : byte;
  Font  : byte;
  Attr  : byte;
//  Value : word;
//  Len   : byte; Text : array of char;
//  Float : double;
 end;

 TBIFF_FORMAT_COUNT=packed record
  Sign  : word;
  Len   : word;
  Count : word;
 end;

 TBIFF_FORMAT=packed record
  Sign  : word;
  Len   : word;
  FmtLen: byte;
  // Format
 end;

 TBIFF_FORMULA=packed record
  Sign   : word;
  Len    : word;
  Row    : word;
  Col    : word;
  Prot   : byte;
  Font   : byte;
  Attr   : byte;
  Value  : double;
  Flag   : byte; // Recalc
  fLen   : byte; // Length(FORMULA)
 // FORMULA
 end;

//----Formula--//

function opAREA(l1,c1,l2,c2:integer):string;
begin
 SetLength(Result,7);
 Result[1]:=#$25;
 Result[2]:=chr(l1 and $ff);
 Result[3]:=chr(((l1 shr 8) and $ff) or $C0);
 Result[4]:=chr(l2 and $ff);
 Result[5]:=chr(((l2 shr 8) and $ff) or $C0);
 Result[6]:=chr(c1);
 Result[7]:=chr(c2);
end;

function opSUM(e:string):string;
begin
 Result:=e+#$19#$10#$00;
end;

function opIF(ACond,AThen,AElse:string):string;
begin
 Result:=
  aCond+#$19#$02+chr(Length(aThen)+3) // Attr, IfControl, aThen+3 bytes from ELSE
 +aThen+#$19#$08+chr(Length(aElse)+5) // Attr, Skip, eElse+5 bytes (+1) from end of IF
 +aElse+#$19#$08#$02 // Attr, Skip, 2 bytes (+1) from end of IF
 +#$42#$03#$01; // FuncVar, 3 args, "IF"
end;

function opEQU(a,b:string):string;
begin
 Result:=a+b+#$0B;
end;

function opREF(l,c:integer):string;
begin
 SetLength(Result,4);
 Result[1]:=#$44;
 Result[2]:=chr(l and $ff);
 Result[3]:=chr(((l shr 8) and $ff) or $C0);
 Result[4]:=chr(c);
end;

function opINT(i:integer):string;
begin
 SetLength(Result,3);
 Result[1]:=#$1E;
 Result[2]:=chr(i and $ff);
 Result[3]:=chr((i shr 8) and $FF);
end;

function opSTR(s:string):string;
var
 l:integer;
begin
 l:=Length(s) and $FF;
 SetLength(Result,l+2);
 Result[1]:=#$17;
 Result[2]:=chr(l);
 if l>0 then Move(s[1],Result[3],l);
end;

function opDIV(a,b:string):string;
begin
 Result:=a+b+#$06;
end;

function opADD(a,b:string):string;
begin
 Result:=a+b+#$03;
end;

function opSUB(a,b:string):string;
begin
 Result:=a+b+#$04;
end;

function opNEG(a:string):string;
begin
 Result:=a+#$13;
end;

function opPAR(e:string):string;
begin
 Result:=e+#$15;
end;

const
 BIFF5:packed array[0..11] of word=(
  $0809,  6, $0200, $0010, $0000,
  $0200, 10, $0000, $FFFF, $0000, $0100, $0000
 );

 BIFF_EOF:packed array[0..1] of word=(
  $000A, 0
 );

procedure SaveFont(Font:TFont; Stream:TStream);
var
 FNT:TBIFF_FONT;
 s:string;
begin
 if Font=nil then exit;
 s:=Font.Name;
 FNT.Sign   :=$0031;
 FNT.Len    :=5+Length(s);
 FNT.Height :=20*Font.Size;
 FNT.Attr   :=byte(Font.Style);
 FNT.NameLen:=Length(s);
 Stream.WriteBuffer(FNT,SizeOf(FNT));
 Stream.WriteBuffer(s[1],FNT.NameLen);
end;

constructor TExcelStream.Create(FileName:string);
begin
 inherited Create(FileName,fmCreate);
 WriteBuffer(BIFF5,SizeOf(BIFF5));
end;

destructor TExcelStream.Destroy;
begin
 WriteBuffer(BIFF_EOF,SizeOf(BIFF_EOF));
 inherited;
end;

procedure TExcelStream.AddFont(Font:TFont);
begin
 SaveFont(Font,Self);
end;

procedure TExcelStream.AddFormats;
const
// # ##0,00\ [$-1]
 Formats:array[0..22] of string=(
  'General',
  '0',
  '0.00',
  '#,##0',
  '#,##0.00',
  '#,##0\ "F";\-#,##0\ "F"',
  '#,##0\ "F";[Red]\-#,##0\ "F"',
  '#,##0.00\ "F";\-#,##0.00\ "F"',
  '0%',
  '0.00%',
  '0.00E+00',
  'dd/mm/yyyy',
  'dd\-mmm\-yy',
  'dd\-mmm',
  'mmm\-yy',
  'h:mm\ AM/PM',
  'h:mm:ss\ AM/PM',
  'hh:mm',
  'hh:mm:ss',
  'dd/mm/yyyy\ hh:mm',
  '##0.0E+0',
  'mm:ss',
  '@'
 );
var
 FMC:TBIFF_FORMAT_COUNT;
 FMT:TBIFF_FORMAT;
 i:integer;
 s:string;
begin
 FMC.Sign :=$001F;
 FMC.Len  :=2;
 FMC.Count:=Length(Formats);
 WriteBuffer(FMC,SizeOf(FMC));
 FMT.Sign:=$001E;
 for i:=low(Formats) to high(Formats) do begin
  s:=Formats[i];
  FMT.Len:=Length(s)+1;
  FMT.FmtLen:=Length(s);
  WriteBuffer(FMT,SizeOf(FMT));
  WriteBuffer(s[1],FMT.FmtLen);
 end;
end;

procedure TExcelStream.SetColWidth(FirstCol, LastCol, Width:word);
var
 CW:TBIFF_COLWIDTH;
begin
 CW.Sign:=$0024;
 CW.Len :=4;
  CW.FirstCol:=FirstCol;
  CW.LastCol :=LastCol;
  CW.Width   :=40*Width;
 WriteBuffer(CW,SizeOf(CW));
end;

procedure TExcelStream.AddInteger(x,y,Value:integer);
var
 CLL:TBIFF_CELL;
begin
 if (Value<0)or(Value>$FFFF) then begin
  AddFloat(x,y,Value);
  exit;
 end;
 CLL.Sign:=2;
 CLL.Len:=9;
  CLL.Row:=y;
  CLL.Col:=x;
  CLL.Prot:=0;
  CLL.Font:=$40*Font;
  CLL.Attr:=ord(fAlignment)+byte(fBorders) shl 3;
 WriteBuffer(CLL,SizeOf(CLL));
 WriteBuffer(Value,2); // WORD !
end;

procedure TExcelStream.AddFloat(x,y:integer; Value:double);
var
 CLL:TBIFF_CELL;
begin
 CLL.Sign:=3;
 CLL.Len:=15;
  CLL.Row:=y;
  CLL.Col:=x;
  CLL.Prot:=0;
  CLL.Font:=$40*Font;
  CLL.Attr:=ord(fAlignment)+byte(fBorders) shl 3;
 WriteBuffer(CLL,SizeOf(CLL));
 WriteBuffer(Value,SizeOf(Value));
end;

procedure TExcelStream.AddCurrency(x,y:integer; Value:double);
var
 CLL:TBIFF_CELL;
begin
 CLL.Sign:=3;
 CLL.Len:=15;
  CLL.Row:=y;
  CLL.Col:=x;
  CLL.Prot:=0;
  CLL.Font:=$40*Font+7 { Currency };
  CLL.Attr:=ord(fAlignment)+byte(fBorders) shl 3;
 WriteBuffer(CLL,SizeOf(CLL));
 WriteBuffer(Value,SizeOf(Value));
end;

procedure TExcelStream.AddString(x,y:integer; Value:string);
var
 CLL:TBIFF_CELL;
 Len:byte;
begin
 Len:=Length(Value);
 if Value='' then begin
  CLL.Sign:=1;
  CLL.Len:=7;
 end else begin
  CLL.Sign:=4;
  CLL.Len:=8+Len;
 end;
  CLL.Row:=y;
  CLL.Col:=x;
  CLL.Prot:=0;
  CLL.Font:=$40*Font;
  CLL.Attr:=ord(fAlignment)+byte(fBorders) shl 3;
 WriteBuffer(CLL,SizeOf(CLL));
 if CLL.Sign=4 then begin
  WriteBuffer(len,SizeOf(len));
  WriteBuffer(Value[1],len);
 end;
end;


procedure TExcelStream.AddSum(x, y, c1, l1, c2, l2: integer; fmt:integer=0);
var
 SUM:TBIFF_FORMULA;
 FORMULA:string;
begin
 FORMULA:=opSUM(opAREA(l1,c1,l2,c2));
 FillChar(SUM,SizeOf(SUM),0);
 SUM.Sign:=6;
 SUM.Len:=SizeOf(SUM)-4+Length(FORMULA);
  SUM.Row:=y;
  SUM.Col:=x;
  SUM.Attr:=ord(fAlignment)+byte(fBorders) shl 3;
  SUM.Font :=$40*Font+fmt;
  SUM.Value:=0;
  SUM.Flag :=1;     // Recalc
  SUM.fLen :=Length(FORMULA);
 WriteBuffer(SUM,SizeOf(SUM));
 WriteBuffer(FORMULA[1],SUM.fLen);
end;

procedure TExcelStream.AddSumDiv2(x, y, c1, l1, c2, l2: integer; fmt:integer=0);
var
 SUM:TBIFF_FORMULA;
 FORMULA:string;
begin
 FORMULA:=opDIV(opSUM(opAREA(l1,c1,l2,c2)),opINT(2));
 FillChar(SUM,SizeOf(SUM),0);
 SUM.Sign:=6;
 SUM.Len:=SizeOf(SUM)-4+Length(FORMULA);
  SUM.Row:=y;
  SUM.Col:=x;
  SUM.Attr:=ord(fAlignment)+byte(fBorders) shl 3;
  SUM.Font :=$40*Font+fmt;
  SUM.Value:=0;
  SUM.Flag :=1;     // Recalc
  SUM.fLen :=Length(FORMULA);
 WriteBuffer(SUM,SizeOf(SUM));
 WriteBuffer(FORMULA[1],SUM.fLen);
end;

procedure TExcelStream.AddSumOfLines(x,y:integer; cells:array of integer; fmt:integer=0);
var
 SUM:TBIFF_FORMULA;
 FORMULA:string;
 i:integer;
 l:integer;
begin
 FORMULA:='';
 for i:=0 to Length(cells)-1 do begin
  l:=cells[i];
  if l>=0 then begin
   if FORMULA='' then FORMULA:=opREF(l,x) else FORMULA:=opADD(FORMULA,opREF(l,x));
  end;
 end;
 if FORMULA='' then
  FORMULA:=opINT(0)
 else begin
  if fmt<0 then begin
   FORMULA:=opNEG(opPAR(FORMULA));
   fmt:=-fmt;
  end;
 end;
 FillChar(SUM,SizeOf(SUM),0);
 SUM.Sign:=6;
 SUM.Len:=SizeOf(SUM)-4+Length(FORMULA);
  SUM.Row:=y;
  SUM.Col:=x;
  SUM.Attr:=ord(fAlignment)+byte(fBorders) shl 3;
  SUM.Font :=$40*Font+fmt;
  SUM.Value:=0;
  SUM.Flag :=1;     // Recalc
  SUM.fLen :=Length(FORMULA);
 WriteBuffer(SUM,SizeOf(SUM));
 WriteBuffer(FORMULA[1],SUM.fLen);
end;

procedure TExcelStream.AddSubLines(x,y,y1,y2:integer; fmt:integer=0);
var
 SUM:TBIFF_FORMULA;
 FORMULA:string;
 i:integer;
 l:integer;
begin
 FORMULA:=opSUB(opREF(y1,x),opREF(y2,x));
 FillChar(SUM,SizeOf(SUM),0);
 SUM.Sign:=6;
 SUM.Len:=SizeOf(SUM)-4+Length(FORMULA);
  SUM.Row:=y;
  SUM.Col:=x;
  SUM.Attr:=ord(fAlignment)+byte(fBorders) shl 3;
  SUM.Font :=$40*Font+fmt;
  SUM.Value:=0;
  SUM.Flag :=1;     // Recalc
  SUM.fLen :=Length(FORMULA);
 WriteBuffer(SUM,SizeOf(SUM));
 WriteBuffer(FORMULA[1],SUM.fLen);
end;

procedure TExcelStream.AddPercent(x, y, c1, l1, c2, l2: integer);
var
 PER:TBIFF_FORMULA;
 FORMULA:string;
begin
 FORMULA:=opIF(
  { IF }opEQU(opREF(l2,c2),opINT(0)),
  {THEN}opSTR(''),
  {ELSE}opDIV(opPAR(opSUB(opREF(l1,c1),opREF(l2,c2))),opREF(l2,c2))
 );
 FillChar(PER,SizeOf(PER),0);
 PER.Sign:=6;
 PER.Len:=SizeOf(PER)-4+Length(FORMULA);
  PER.Row:=y;
  PER.Col:=x;
  PER.Attr:=ord(fAlignment)+byte(fBorders) shl 3;
  PER.Font :=$40*Font + 10 { 0.00% };
  PER.Value:=0;
  PER.Flag :=1;     // Recalc
  PER.fLen :=Length(FORMULA);
 WriteBuffer(PER,SizeOf(PER));
 WriteBuffer(FORMULA[1],PER.fLen);
end;

procedure TExcelStream.AddRef(x, y, c1, l1: integer; fmt:integer=0);
var
 PER:TBIFF_FORMULA;
 FORMULA:string;
begin
 FORMULA:=opREF(l1,c1);
 FillChar(PER,SizeOf(PER),0);
 PER.Sign:=6;
 PER.Len:=SizeOf(PER)-4+Length(FORMULA);
  PER.Row:=y;
  PER.Col:=x;
  PER.Attr:=ord(fAlignment)+byte(fBorders) shl 3;
  PER.Font :=$40*Font + fmt;
  PER.Value:=0;
  PER.Flag :=1;     // Recalc
  PER.fLen :=Length(FORMULA);
 WriteBuffer(PER,SizeOf(PER));
 WriteBuffer(FORMULA[1],PER.fLen);
end;

function TExcelStream.GetFont: integer;
begin
 if fFont>4 then Result:=fFont-1 else Result:=fFont;
end;

procedure TExcelStream.SetFont(Value: integer);
begin
 if Value>=4 then fFont:=Value+1 else fFont:=Value;
end;

end.
